"""Create dash plot for /plot endpoint."""

from __future__ import annotations
import os
import pandas as pd
import plotly.graph_objects as go
import plotly.express as px
import plotly.io as pio
from dash import Dash
from dash.dependencies import Input, Output
import dash_core_components as dcc
import dash_html_components as html
from bairy.device import configs, preprocess


pio.templates.default = 'plotly_white'


def create_fig(time_period: str = 'all'):
  """Create plotly figure using one or two y-axes."""

  data_path = configs.PREPROCESSED_DATA_PATHS[time_period]
  # avoiding errors before any data is captured
  if not os.path.exists(data_path):
    return go.Figure()

  fig = go.Figure()
  # see https://plotly.com/python/discrete-color/
  colors = iter(px.colors.qualitative.Bold)
  df = pd.read_csv(data_path)
  sensor_headers, sensor_units = preprocess.determine_plot_configs()

  for i, (key, cols) in enumerate(sensor_headers.items()):
    unit = sensor_units[key]
    if len(sensor_headers) == 2 and i == 0:
      side = 'right'
      opacity = 0.3
    else:
      side = 'left'
      opacity = 1.0

    yaxis = 'y' if i == 0 else 'y2'
    yaxis_layout = 'yaxis' if i == 0 else 'yaxis2'
    layout_params = {yaxis_layout: {'title': unit, 'side': side}}
    if i == 1:
      layout_params[yaxis_layout]['overlaying'] = 'y'

    for col in cols:
      # making some traces traces hidden
      if col in ['random3', 'pm_1.0', 'motion', 'sound']:
        visible = 'legendonly'
      else:
        visible = None

      fig.add_trace(go.Scatter(
          x=df['time'],
          y=df[col],
          name=col,
          opacity=opacity,
          line={'color': next(colors)},
          yaxis=yaxis,
          visible=visible
      ))

    fig.update_layout(**layout_params)

  # threshold for pm2.5
  if 'air' in sensor_headers and df['pm_2.5'].max() > 40:
    yaxis = 'y2' if len(sensor_headers) == 2 else 'y'
    fig.add_trace(go.Scatter(
        x=df['time'],
        y=[35] * len(df),
        name='pm_2.5 safe threshold',
        yaxis=yaxis,
        line={'dash': 'dot', 'color': next(colors)}
    ))

  name = configs.load_device().name
  if time_period == 'day':
    title = f'{name} data over last 24 hours'
  elif time_period == 'week':
    title = f'{name} data over last 7 days'
  else:
    title = f'{name} data over entire runtime'

  fig.update_layout(height=800, title=title)
  fig.layout.xaxis.rangeslider.visible = True
  fig.layout.yaxis.fixedrange = False
  return fig


css = 'https://codepen.io/chriddyp/pen/bWLwgP.css'
plot = Dash(
    __name__,
    requests_pathname_prefix='/plot/',
    external_stylesheets=[css]
)
plot.layout = html.Div(children=[
    # title and links
    html.Div(children=[
        html.H1(children='bairy', style={'fontWeight': 'bold'}),
        html.H3(children='Display sensor data from Raspberry Pi.'),
        html.A(
            'about bairy',
            href='https://github.com/zebengberg/bairy',
            style={'margin': '20px'}
        ),
        html.A(
            'table',
            href='/table',
            style={'margin': '20px'}
        )
    ]),

    # three plots
    dcc.Graph(id='plot_day'),
    dcc.Graph(id='plot_week'),
    dcc.Graph(id='plot_all'),

    # for plot update callback
    # see https: // dash.plotly.com/live-updates
    dcc.Interval(
        id='interval-component',
        interval=5 * 60 * 1000,  # 5 minutes
    )
])


@plot.callback(
    Output('plot_day', 'figure'),
    Input('interval-component', 'n_intervals')
)
def serve_plot_day(_):
  """Dynamically serve dash_plot.layout."""
  return create_fig('day')


@plot.callback(
    Output('plot_week', 'figure'),
    Input('interval-component', 'n_intervals')
)
def serve_plot_week(_):
  """Dynamically serve dash_plot.layout."""
  return create_fig('week')


@plot.callback(
    Output('plot_all', 'figure'),
    Input('interval-component', 'n_intervals')
)
def serve_plot_all(_):
  """Dynamically serve dash_plot.layout."""
  return create_fig('all')



"""Dash app to plot device data."""

import os
import glob
import pandas as pd
import plotly.graph_objects as go
import plotly.express as px
from dash import Dash
import dash_core_components as dcc
import dash_html_components as html
from bairy.hub import configs


def load_data(time_period: str = 'all'):
  """Load cached data."""

  dfs = glob.glob(configs.HUB_DATA_DIR + '/*' + time_period + '.csv')
  names = [f.split('/')[-1].split('_')[0] for f in dfs]

  dfs = [pd.read_csv(f) for f in dfs]

  # must include time
  cols_to_keep = {'time', 'pm_10', 'pm_2.5', 'random1'}
  dfs = [df[cols_to_keep & set(df.columns)] for df in dfs]

  dfs = [df.set_index(pd.to_datetime(df['time'])) for df in dfs]
  dfs = [df.drop('time', axis=1) for df in dfs]
  dfs = [df.rename(columns={k: name + ' ' + k for k in df.columns})
         for df, name in zip(dfs, names)]
  if dfs == []:
    return pd.DataFrame([])
  return pd.concat(dfs)


def create_fig(time_period: str):
  """Create plotly figure using one or two y-axes."""

  # avoiding errors when device not configured to run as hub
  if not os.path.exists(configs.IP_PATH):
    return px.line()

  df = load_data(time_period)
  if df.empty:
    return px.line()

  fig = px.line(df, x=df.index, y=df.columns)

  # showing pm2.5 safe threshold in any column exceeds
  for k in df.columns:
    if 'pm_2.5' in k:
      if df[k].max() > 40:
        fig.add_trace(go.Scatter(
            x=df.index,
            y=[35] * len(df),
            name='pm_2.5 safe threshold',
            line={'dash': 'dot'}))
        break

  if time_period == 'day':
    title = 'data over last 24 hours'
  elif time_period == 'week':
    title = 'data over last 7 days'
  else:
    title = 'data over entire runtime'

  fig.update_layout(height=800, title=title)
  fig.layout.xaxis.rangeslider.visible = True
  fig.layout.yaxis.title = 'micrograms / cubic meter'
  return fig


def serve_plot():
  """Serve dash app plot."""
  fig_day = create_fig('day')
  fig_week = create_fig('week')

  return html.Div(children=[
      html.Div(children=[
          html.H1(children='bairy', style={'fontWeight': 'bold'}),
          html.H3(children='Display sensor data from Raspberry Pi.'),
          html.A(
              'about bairy',
              href='https://github.com/zebengberg/bairy',
              style={'margin': '20px'})]),
      dcc.Graph(id='graph_day', figure=fig_day),
      dcc.Graph(id='graph_week', figure=fig_week)])


css = 'https://codepen.io/chriddyp/pen/bWLwgP.css'
dash_plot = Dash(
    requests_pathname_prefix='/plot/',
    external_stylesheets=[css]
)
dash_plot.layout = serve_plot



"""Create dash table for /table endpoint."""

import os
import pandas as pd
from dash import Dash
from dash.dependencies import Input, Output
from dash_table import DataTable
import dash_core_components as dcc
import dash_html_components as html
from bairy.device.dash_plot import css
from bairy.device import configs

table = Dash(
    requests_pathname_prefix='/table/',
    external_stylesheets=[css]
)
table.layout = html.Div(children=[
    # title and links
    html.Div(children=[
        html.H1(children='bairy', style={'fontWeight': 'bold'}),
        html.H3(children='Display sensor data from Raspberry Pi.'),
        html.A(
            'about bairy',
            href='https://github.com/zebengberg/bairy',
            style={'margin': '20px'}
        ),
        html.A(
            'plot',
            href='/plot',
            style={'margin': '20px'}
        )
    ]),

    # table
    DataTable(
        id='table',
        style_cell=dict(textAlign='left'),
        style_header=dict(backgroundColor="paleturquoise"),
        style_data=dict(backgroundColor="lavender")
    ),

    # for plot update callback
    # see https://dash.plotly.com/live-updates
    dcc.Interval(
        id='interval-component',
        interval=60 * 1000,  # in milliseconds
        n_intervals=0  # an unused counter
    )
])


@table.callback(
    [Output('table', 'columns'), Output('table', 'data')],
    Input('interval-component', 'n_intervals')
)
def serve_table(_):
  """Dynamically serve table columns and data."""

  data_path = configs.PREPROCESSED_DATA_PATHS['day']
  # avoiding errors before any data is captured
  if not os.path.exists(data_path):
    return None, None

  df = pd.read_csv(data_path, index_col=0)
  df = df.iloc[::-1]
  df.index = df.index.astype(str)
  df = df.round(3)
  columns = [{'name': i, 'id': i} for i in df.columns]
  data = df.to_dict('records')
  return columns, data



"""Preprocess data for Plotly/Dash."""

from __future__ import annotations
import itertools
import asyncio
import pandas as pd
from bairy.device import configs


def determine_plot_configs():
  """Determine plotly axes and variables configurations."""

  d = configs.load_device()
  # the order of the dictionaries below matter for plot tracing
  # order is least important to most important
  sensor_headers: dict[str, list[str]] = {'random': [],
                                          'digital': [],
                                          'air': []}
  sensor_units = {'random': 'random',
                  'digital': 'intensity',
                  'air': 'micrograms / cubic meter'}

  for s in d.sensors:
    if s.sensor_type == 'air':
      sensor_headers['air'] += ['pm_1.0', 'pm_2.5', 'pm_10']
    else:
      sensor_headers[s.sensor_type].append(s.header)

  # cleanup dictionaries
  empties = [k for k in sensor_headers.keys() if sensor_headers[k] == []]
  for k in empties:
    del sensor_headers[k]
    del sensor_units[k]

  # keeping at most two types of headers
  if len(sensor_headers) == 3:
    del sensor_headers['random']
    del sensor_units['random']

  return sensor_headers, sensor_units


def preprocess_df(time_period: str = 'all'):
  """Preprocess pandas DataFrame."""
  df = pd.read_csv(configs.DATA_PATH)

  df = df.set_index('time')
  df.index = pd.to_datetime(df.index)

  sensor_headers, _ = determine_plot_configs()
  cols_to_keep = [col for key in sensor_headers for col in sensor_headers[key]]
  df = df[cols_to_keep]

  if time_period == 'day':
    start = pd.Timestamp.now() - pd.Timedelta('1 day')
    df = df[df.index > start]
  elif time_period == 'week':
    start = pd.Timestamp.now() - pd.Timedelta('7 days')
    df = df[df.index > start]

  df = resample_df(df)
  return df.reset_index()  # move time back as a column


def resample_df(df):
  """Smooth and condense df by resampling."""
  # 60 has many divisors
  rules = ['1T', '2T', '3T', '4T', '5T', '6T', '10T', '20T', '30T', '1H']

  if len(df) > 300:  # only resample if df large enough
    for r in rules:
      df = df.resample(r).mean()
      if len(df) < 5000:
        break

  # smoothing even more
  df = df.rolling(7, center=True, min_periods=1).mean()
  return df


async def run_preprocess():
  """Run preprocessing indefinitely."""

  time_periods = list(configs.PREPROCESSED_DATA_PATHS.keys())

  async def run():
    for time_period in itertools.cycle(time_periods):
      df = preprocess_df(time_period)
      df.to_csv(configs.PREPROCESSED_DATA_PATHS[time_period])
      await asyncio.sleep(60)

  return await asyncio.create_task(run())
